@@ -9,8 +9,10 @@
 #include <config.h>
 #endif
 
+#include "soup-address.h"
 #include "soup-session-async.h"
 #include "soup-session-private.h"
+#include "soup-address.h"
 #include "soup-message-private.h"
 #include "soup-misc.h"
 
@@ -24,7 +26,7 @@
  **/
 
 static gboolean run_queue (SoupSessionAsync *sa);
-static void do_idle_run_queue (SoupSessionAsync *sa);
+static void do_idle_run_queue (SoupSession *session);
 
 static void  queue_message   (SoupSession *session, SoupMessage *req,
 			      SoupSessionCallback callback, gpointer user_data);
@@ -108,22 +110,107 @@ soup_session_async_new_with_options (const char *optname1, ...)
 
 
 static void
-connection_closed (SoupConnection *conn, gpointer sa)
+resolved_msg_addr (SoupAddress *addr, guint status, gpointer user_data)
+{
+	SoupMessageQueueItem *item = user_data;
+	SoupSession *session = item->session;
+
+	if (item->removed) {
+		/* Message was cancelled before its address resolved */
+		soup_message_queue_item_unref (item);
+		return;
+	}
+
+	if (!SOUP_STATUS_IS_SUCCESSFUL (status)) {
+		soup_session_cancel_message (session, item->msg, status);
+		soup_message_queue_item_unref (item);
+		return;
+	}
+
+	item->msg_addr = g_object_ref (addr);
+	item->resolving_msg_addr = FALSE;
+
+	soup_message_queue_item_unref (item);
+
+	/* If we got here we know session still exists */
+	run_queue ((SoupSessionAsync *)session);
+}
+
+static void
+resolve_msg_addr (SoupMessageQueueItem *item)
+{
+	if (item->resolving_msg_addr)
+		return;
+	item->resolving_msg_addr = TRUE;
+
+	soup_message_queue_item_ref (item);
+	soup_address_resolve_async (soup_message_get_address (item->msg),
+				    soup_session_get_async_context (item->session),
+				    item->cancellable,
+				    resolved_msg_addr, item);
+}
+
+static void
+resolved_proxy_addr (SoupProxyResolver *proxy_resolver, SoupMessage *msg,
+		     guint status, SoupAddress *proxy_addr, gpointer user_data)
+{
+	SoupMessageQueueItem *item = user_data;
+	SoupSession *session = item->session;
+
+	if (item->removed) {
+		/* Message was cancelled before its proxy addr resolved */
+		soup_message_queue_item_unref (item);
+		return;
+	}
+
+	if (!SOUP_STATUS_IS_SUCCESSFUL (status)) {
+		soup_session_cancel_message (session, item->msg, status);
+		soup_message_queue_item_unref (item);
+		return;
+	}
+
+	item->resolving_proxy_addr = FALSE;
+	item->resolved_proxy_addr = TRUE;
+	item->proxy_addr = proxy_addr ? g_object_ref (proxy_addr) : NULL;
+
+	soup_message_queue_item_unref (item);
+
+	/* If we got here we know session still exists */
+	run_queue ((SoupSessionAsync *)session);
+}
+
+static void
+resolve_proxy_addr (SoupMessageQueueItem *item,
+		    SoupProxyResolver *proxy_resolver)
+{
+	if (item->resolving_proxy_addr)
+		return;
+	item->resolving_proxy_addr = TRUE;
+
+	soup_message_queue_item_ref (item);
+	soup_proxy_resolver_get_proxy_async (proxy_resolver, item->msg,
+					     soup_session_get_async_context (item->session),
+					     item->cancellable,
+					     resolved_proxy_addr, item);
+}
+
+static void
+connection_closed (SoupConnection *conn, gpointer session)
 {
 	/* Run the queue in case anyone was waiting for a connection
 	 * to be closed.
 	 */
-	do_idle_run_queue (sa);
+	do_idle_run_queue (session);
 }
 
 static void
 got_connection (SoupConnection *conn, guint status, gpointer user_data)
 {
-	SoupSessionAsync *sa = user_data;
+	SoupSession *session = user_data;
 
 	if (status == SOUP_STATUS_OK) {
 		g_signal_connect (conn, "disconnected",
-				  G_CALLBACK (connection_closed), sa);
+				  G_CALLBACK (connection_closed), session);
 
 		/* @conn has been marked reserved by SoupSession, but
 		 * we don't actually have any specific message in mind
@@ -142,8 +229,8 @@ got_connection (SoupConnection *conn, guint status, gpointer user_data)
 	 * there may have been messages waiting for the connection
 	 * count to go down.
 	 */
-	do_idle_run_queue (sa);
-	g_object_unref (sa);
+	do_idle_run_queue (session);
+	g_object_unref (session);
 }
 
 static gboolean
@@ -151,7 +238,9 @@ run_queue (SoupSessionAsync *sa)
 {
 	SoupSession *session = SOUP_SESSION (sa);
 	SoupMessageQueue *queue = soup_session_get_queue (session);
-	SoupMessageQueueIter iter;
+	SoupMessageQueueItem *item;
+	SoupProxyResolver *proxy_resolver =
+		soup_session_get_proxy_resolver (session);
 	SoupMessage *msg;
 	SoupConnection *conn;
 	gboolean try_pruning = TRUE, should_prune = FALSE;
@@ -160,15 +249,26 @@ run_queue (SoupSessionAsync *sa)
 	/* FIXME: prefer CONNECTING messages */
 
  try_again:
-	for (msg = soup_message_queue_first (queue, &iter);
-	     msg && !should_prune;
-	     msg = soup_message_queue_next (queue, &iter)) {
+	for (item = soup_message_queue_first (queue);
+	     item && !should_prune;
+	     item = soup_message_queue_next (queue, item)) {
+		msg = item->msg;
 
 		if (!SOUP_MESSAGE_IS_STARTING (msg) ||
 		    soup_message_io_in_progress (msg))
 			continue;
 
+		if (!item->msg_addr) {
+			resolve_msg_addr (item);
+			continue;
+		}
+		if (proxy_resolver && !item->resolved_proxy_addr) {
+			resolve_proxy_addr (item, proxy_resolver);
+			continue;
+		}
+
 		conn = soup_session_get_connection (session, msg,
+						    item->proxy_addr,
 						    &should_prune, &is_new);
 		if (!conn)
 			continue;
@@ -179,6 +279,8 @@ run_queue (SoupSessionAsync *sa)
 		} else
 			soup_connection_send_request (conn, msg);
 	}
+	if (item)
+		soup_message_queue_item_unref (item);
 
 	if (try_pruning && should_prune) {
 		/* There is at least one message in the queue that
@@ -186,7 +288,7 @@ run_queue (SoupSessionAsync *sa)
 		 * some other server.
 		 */
 		if (soup_session_try_prune_connection (session)) {
-			try_pruning = FALSE;
+			try_pruning = should_prune = FALSE;
 			goto try_again;
 		}
 	}
@@ -195,37 +297,43 @@ run_queue (SoupSessionAsync *sa)
 }
 
 static void
-request_restarted (SoupMessage *req, gpointer sa)
+request_restarted (SoupMessage *req, gpointer user_data)
 {
-	run_queue (sa);
-}
+	SoupMessageQueueItem *item = user_data;
 
-typedef struct {
-	SoupSessionAsync *sa;
-	SoupSessionCallback callback;
-	gpointer callback_data;
-} SoupSessionAsyncQueueData;
+	if (item->msg_addr &&
+	    item->msg_addr != soup_message_get_address (item->msg)) {
+		g_object_unref (item->msg_addr);
+		item->msg_addr = NULL;
+	}
+	if (item->proxy_addr) {
+		g_object_unref (item->proxy_addr);
+		item->proxy_addr = NULL;
+	}
+	item->resolved_proxy_addr = FALSE;
+
+	run_queue ((SoupSessionAsync *)item->session);
+}
 
 static void
 final_finished (SoupMessage *req, gpointer user_data)
 {
-	SoupSessionAsyncQueueData *saqd = user_data;
-	SoupSessionAsync *sa = saqd->sa;
+	SoupMessageQueueItem *item = user_data;
+	SoupSession *session = item->session;
+
+	g_object_ref (session);
 
-	g_object_ref (sa);
 	if (!SOUP_MESSAGE_IS_STARTING (req)) {
-		g_signal_handlers_disconnect_by_func (req, final_finished, saqd);
-		if (saqd->callback) {
-			saqd->callback ((SoupSession *)sa, req,
-					saqd->callback_data);
-		}
+		g_signal_handlers_disconnect_by_func (req, final_finished, item);
+		if (item->callback)
+			item->callback (session, req, item->callback_data);
 
 		g_object_unref (req);
-		g_slice_free (SoupSessionAsyncQueueData, saqd);
+		soup_message_queue_item_unref (item);
 	}
 
-	do_idle_run_queue (sa);
-	g_object_unref (sa);
+	do_idle_run_queue (session);
+	g_object_unref (session);
 }
 
 static gboolean
@@ -239,14 +347,14 @@ idle_run_queue (gpointer sa)
 }
 
 static void
-do_idle_run_queue (SoupSessionAsync *sa)
+do_idle_run_queue (SoupSession *session)
 {
-	SoupSessionAsyncPrivate *priv = SOUP_SESSION_ASYNC_GET_PRIVATE (sa);
+	SoupSessionAsyncPrivate *priv = SOUP_SESSION_ASYNC_GET_PRIVATE (session);
 
 	if (!priv->idle_run_queue_source) {
 		priv->idle_run_queue_source = soup_add_completion (
-			soup_session_get_async_context ((SoupSession *)sa),
-			idle_run_queue, sa);
+			soup_session_get_async_context (session),
+			idle_run_queue, session);
 	}
 }
 
@@ -254,21 +362,19 @@ static void
 queue_message (SoupSession *session, SoupMessage *req,
 	       SoupSessionCallback callback, gpointer user_data)
 {
-	SoupSessionAsync *sa = SOUP_SESSION_ASYNC (session);
-	SoupSessionAsyncQueueData *saqd;
-
-	g_signal_connect (req, "restarted",
-			  G_CALLBACK (request_restarted), sa);
-
-	saqd = g_slice_new (SoupSessionAsyncQueueData);
-	saqd->sa = sa;
-	saqd->callback = callback;
-	saqd->callback_data = user_data;
-	g_signal_connect_after (req, "finished",
-				G_CALLBACK (final_finished), saqd);
+	SoupMessageQueueItem *item;
 
 	SOUP_SESSION_CLASS (soup_session_async_parent_class)->queue_message (session, req, callback, user_data);
-	do_idle_run_queue (sa);
+
+	item = soup_message_queue_lookup (soup_session_get_queue (session), req);
+	g_return_if_fail (item != NULL);
+
+	g_signal_connect (req, "restarted",
+			  G_CALLBACK (request_restarted), item);
+	g_signal_connect_after (req, "finished",
+				G_CALLBACK (final_finished), item);
+
+	do_idle_run_queue (session);
 }
 
 static guint
