@@ -13,6 +13,7 @@
 #include <string.h>
 #include <stdlib.h>
 
+#include "soup-address.h"
 #include "soup-auth.h"
 #include "soup-auth-basic.h"
 #include "soup-auth-digest.h"
@@ -21,6 +22,7 @@
 #include "soup-marshal.h"
 #include "soup-message-private.h"
 #include "soup-message-queue.h"
+#include "soup-proxy-resolver-static.h"
 #include "soup-session.h"
 #include "soup-session-feature.h"
 #include "soup-session-private.h"
@@ -55,18 +57,14 @@
  **/
 
 typedef struct {
-	SoupURI    *root_uri;
+	SoupAddress *addr;
 
-	GSList     *connections;      /* CONTAINS: SoupConnection */
-	guint       num_conns;
-
-	GHashTable *auth_realms;      /* path -> scheme:realm */
-	GHashTable *auths;            /* scheme:realm -> SoupAuth */
+	GSList      *connections;      /* CONTAINS: SoupConnection */
+	guint        num_conns;
 } SoupSessionHost;
 
 typedef struct {
-	SoupURI *proxy_uri;
-	SoupAuth *proxy_auth;
+	SoupProxyResolver *proxy_resolver;
 
 	char *ssl_ca_file;
 	SoupSSLCredentials *ssl_creds;
@@ -78,7 +76,7 @@ typedef struct {
 	GSList *features;
 	SoupAuthManager *auth_manager;
 
-	GHashTable *hosts; /* SoupURI -> SoupSessionHost */
+	GHashTable *hosts; /* SoupAddress -> SoupSessionHost */
 	GHashTable *conns; /* SoupConnection -> SoupSessionHost */
 	guint num_conns;
 	guint max_conns, max_conns_per_host;
@@ -106,12 +104,6 @@ static void auth_manager_authenticate (SoupAuthManager *manager,
 				       SoupMessage *msg, SoupAuth *auth,
 				       gboolean retrying, gpointer user_data);
 
-/* temporary until we fix this to index hosts by SoupAddress */
-extern guint     soup_uri_host_hash  (gconstpointer  key);
-extern gboolean  soup_uri_host_equal (gconstpointer  v1,
-				      gconstpointer  v2);
-extern SoupURI  *soup_uri_copy_root  (SoupURI *uri);
-
 #define SOUP_SESSION_MAX_CONNS_DEFAULT 10
 #define SOUP_SESSION_MAX_CONNS_PER_HOST_DEFAULT 2
 
@@ -158,11 +150,11 @@ soup_session_init (SoupSession *session)
 {
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
 
-	priv->queue = soup_message_queue_new ();
+	priv->queue = soup_message_queue_new (session);
 
 	priv->host_lock = g_mutex_new ();
-	priv->hosts = g_hash_table_new (soup_uri_host_hash,
-					soup_uri_host_equal);
+	priv->hosts = g_hash_table_new (soup_address_hash_by_ip,
+					soup_address_equal_by_ip);
 	priv->conns = g_hash_table_new (NULL, NULL);
 
 	priv->max_conns = SOUP_SESSION_MAX_CONNS_DEFAULT;
@@ -192,8 +184,8 @@ cleanup_hosts (SoupSessionPrivate *priv)
 
 	g_mutex_lock (priv->host_lock);
 	old_hosts = priv->hosts;
-	priv->hosts = g_hash_table_new (soup_uri_host_hash,
-					soup_uri_host_equal);
+	priv->hosts = g_hash_table_new (soup_address_hash_by_ip,
+					soup_address_equal_by_ip);
 	g_mutex_unlock (priv->host_lock);
 
 	g_hash_table_foreach_remove (old_hosts, foreach_free_host, NULL);
@@ -232,9 +224,6 @@ finalize (GObject *object)
 	if (priv->auth_manager)
 		g_object_unref (priv->auth_manager);
 
-	if (priv->proxy_uri)
-		soup_uri_free (priv->proxy_uri);
-
 	if (priv->ssl_creds)
 		soup_ssl_free_client_credentials (priv->ssl_creds);
 
@@ -510,18 +499,6 @@ soup_session_class_init (SoupSessionClass *session_class)
 				    G_PARAM_READWRITE));
 }
 
-static gboolean
-safe_uri_equal (SoupURI *a, SoupURI *b)
-{
-	if (!a && !b)
-		return TRUE;
-
-	if ((a && !b) || (b && !a))
-		return FALSE;
-
-	return soup_uri_equal (a, b);
-}
-
 static gboolean
 safe_str_equal (const char *a, const char *b)
 {
@@ -541,7 +518,6 @@ set_property (GObject *object, guint prop_id,
 	SoupSession *session = SOUP_SESSION (object);
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
 	SoupURI *uri;
-	gboolean need_abort = FALSE;
 	gboolean ca_file_changed = FALSE;
 	const char *new_ca_file, *user_agent;
 
@@ -549,19 +525,16 @@ set_property (GObject *object, guint prop_id,
 	case PROP_PROXY_URI:
 		uri = g_value_get_boxed (value);
 
-		if (!safe_uri_equal (priv->proxy_uri, uri))
-			need_abort = TRUE;
-
-		if (priv->proxy_uri)
-			soup_uri_free (priv->proxy_uri);
-
-		priv->proxy_uri = uri ? soup_uri_copy (uri) : NULL;
-
-		if (need_abort) {
-			soup_session_abort (session);
-			cleanup_hosts (priv);
-		}
+		if (uri) {
+			soup_session_remove_feature_by_type (session, SOUP_TYPE_PROXY_RESOLVER);
+			priv->proxy_resolver = soup_proxy_resolver_static_new (uri);
+			soup_session_add_feature (session, SOUP_SESSION_FEATURE (priv->proxy_resolver));
+			g_object_unref (priv->proxy_resolver);
+		} else if (priv->proxy_resolver && SOUP_IS_PROXY_RESOLVER_STATIC (priv->proxy_resolver))
+			soup_session_remove_feature_by_type (session, SOUP_TYPE_PROXY_RESOLVER);
 
+		soup_session_abort (session);
+		cleanup_hosts (priv);
 		break;
 	case PROP_MAX_CONNS:
 		priv->max_conns = g_value_get_int (value);
@@ -643,7 +616,12 @@ get_property (GObject *object, guint prop_id,
 
 	switch (prop_id) {
 	case PROP_PROXY_URI:
-		g_value_set_boxed (value, priv->proxy_uri);
+		if (priv->proxy_resolver && SOUP_IS_PROXY_RESOLVER_STATIC (priv->proxy_resolver)) {
+			g_object_get_property (G_OBJECT (priv->proxy_resolver),
+					       SOUP_PROXY_RESOLVER_STATIC_PROXY_URI,
+					       value);
+		} else
+			g_value_set_boxed (value, NULL);
 		break;
 	case PROP_MAX_CONNS:
 		g_value_set_int (value, priv->max_conns);
@@ -702,20 +680,12 @@ soup_session_get_async_context (SoupSession *session)
 /* Hosts */
 
 static SoupSessionHost *
-soup_session_host_new (SoupSession *session, SoupURI *source_uri)
+soup_session_host_new (SoupSession *session, SoupAddress *addr)
 {
-	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
 	SoupSessionHost *host;
 
 	host = g_slice_new0 (SoupSessionHost);
-	host->root_uri = soup_uri_copy_root (source_uri);
-
-	if (host->root_uri->scheme == SOUP_URI_SCHEME_HTTPS &&
-	    !priv->ssl_creds) {
-		priv->ssl_creds =
-			soup_ssl_get_client_credentials (priv->ssl_ca_file);
-	}
-
+	host->addr = g_object_ref (addr);
 	return host;
 }
 
@@ -728,14 +698,14 @@ get_host_for_message (SoupSession *session, SoupMessage *msg)
 {
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
 	SoupSessionHost *host;
-	SoupURI *source = soup_message_get_uri (msg);
+	SoupAddress *addr = soup_message_get_address (msg);
 
-	host = g_hash_table_lookup (priv->hosts, source);
+	host = g_hash_table_lookup (priv->hosts, addr);
 	if (host)
 		return host;
 
-	host = soup_session_host_new (session, source);
-	g_hash_table_insert (priv->hosts, host->root_uri, host);
+	host = soup_session_host_new (session, addr);
+	g_hash_table_insert (priv->hosts, host->addr, host);
 
 	return host;
 }
@@ -750,7 +720,7 @@ free_host (SoupSessionHost *host)
 		soup_connection_disconnect (conn);
 	}
 
-	soup_uri_free (host->root_uri);
+	g_object_unref (host->addr);
 	g_slice_free (SoupSessionHost, host);
 }	
 
@@ -920,7 +890,7 @@ connect_result (SoupConnection *conn, guint status, gpointer user_data)
 	SoupSession *session = user_data;
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
 	SoupSessionHost *host;
-	SoupMessageQueueIter iter;
+	SoupMessageQueueItem *item;
 	SoupMessage *msg;
 
 	g_mutex_lock (priv->host_lock);
@@ -960,7 +930,9 @@ connect_result (SoupConnection *conn, guint status, gpointer user_data)
 	 * any messages waiting for this host, since they're out
 	 * of luck.
 	 */
-	for (msg = soup_message_queue_first (priv->queue, &iter); msg; msg = soup_message_queue_next (priv->queue, &iter)) {
+	g_object_ref (session);
+	for (item = soup_message_queue_first (priv->queue); item; item = soup_message_queue_next (priv->queue, item)) {
+		msg = item->msg;
 		if (get_host_for_message (session, msg) == host) {
 			if (status == SOUP_STATUS_TRY_AGAIN) {
 				if (soup_message_get_io_status (msg) == SOUP_MESSAGE_IO_STATUS_CONNECTING)
@@ -971,6 +943,7 @@ connect_result (SoupConnection *conn, guint status, gpointer user_data)
 			}
 		}
 	}
+	g_object_unref (session);
 }
 
 /**
@@ -1013,12 +986,15 @@ connect_result (SoupConnection *conn, guint status, gpointer user_data)
  **/
 SoupConnection *
 soup_session_get_connection (SoupSession *session, SoupMessage *msg,
+			     SoupAddress *proxy_addr,
 			     gboolean *try_pruning, gboolean *is_new)
 {
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
 	SoupConnection *conn;
 	SoupSessionHost *host;
+	SoupSSLCredentials *ssl_creds;
 	GSList *conns;
+	SoupURI *uri;
 
 	g_mutex_lock (priv->host_lock);
 
@@ -1051,10 +1027,18 @@ soup_session_get_connection (SoupSession *session, SoupMessage *msg,
 		return NULL;
 	}
 
+	uri = soup_message_get_uri (msg);
+	if (uri->scheme == SOUP_URI_SCHEME_HTTPS) {
+		if (!priv->ssl_creds)
+			priv->ssl_creds = soup_ssl_get_client_credentials (priv->ssl_ca_file);
+		ssl_creds = priv->ssl_creds;
+	} else
+		ssl_creds = NULL;
+
 	conn = soup_connection_new (
-		SOUP_CONNECTION_ORIGIN_URI, host->root_uri,
-		SOUP_CONNECTION_PROXY_URI, priv->proxy_uri,
-		SOUP_CONNECTION_SSL_CREDENTIALS, priv->ssl_creds,
+		SOUP_CONNECTION_SERVER_ADDRESS, host->addr,
+		SOUP_CONNECTION_PROXY_ADDRESS, proxy_addr,
+		SOUP_CONNECTION_SSL_CREDENTIALS, ssl_creds,
 		SOUP_CONNECTION_ASYNC_CONTEXT, priv->async_context,
 		SOUP_CONNECTION_TIMEOUT, priv->io_timeout,
 		SOUP_CONNECTION_IDLE_TIMEOUT, priv->idle_timeout,
@@ -1099,14 +1083,21 @@ soup_session_get_queue (SoupSession *session)
 static void
 message_finished (SoupMessage *msg, gpointer user_data)
 {
-	SoupSession *session = user_data;
+	SoupMessageQueueItem *item = user_data;
+	SoupSession *session = item->session;
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
 
 	if (!SOUP_MESSAGE_IS_STARTING (msg)) {
-		soup_message_queue_remove_message (priv->queue, msg);
-		g_signal_handlers_disconnect_by_func (msg, message_finished, session);
-		g_signal_handlers_disconnect_by_func (msg, redirect_handler, session);
+		soup_message_queue_remove (priv->queue, item);
+		g_signal_handlers_disconnect_by_func (msg, message_finished, item);
+		/* g_signal_handlers_disconnect_by_func doesn't work if you
+		 * have a metamarshal, meaning it doesn't work with
+		 * soup_message_add_header_handler()
+		 */
+		g_signal_handlers_disconnect_matched (msg, G_SIGNAL_MATCH_DATA,
+						      0, 0, NULL, NULL, session);
 		g_signal_emit (session, signals[REQUEST_UNQUEUED], 0, msg);
+		soup_message_queue_item_unref (item);
 	}
 }
 
@@ -1115,9 +1106,13 @@ queue_message (SoupSession *session, SoupMessage *msg,
 	       SoupSessionCallback callback, gpointer user_data)
 {
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
+	SoupMessageQueueItem *item;
+
+	item = soup_message_queue_append (priv->queue, msg, callback, user_data);
+	soup_message_set_io_status (msg, SOUP_MESSAGE_IO_STATUS_QUEUED);
 
 	g_signal_connect_after (msg, "finished",
-				G_CALLBACK (message_finished), session);
+				G_CALLBACK (message_finished), item);
 
 	if (!(soup_message_get_flags (msg) & SOUP_MESSAGE_NO_REDIRECT)) {
 		soup_message_add_header_handler (
@@ -1125,9 +1120,6 @@ queue_message (SoupSession *session, SoupMessage *msg,
 			G_CALLBACK (redirect_handler), session);
 	}
 
-	soup_message_set_io_status (msg, SOUP_MESSAGE_IO_STATUS_QUEUED);
-	soup_message_queue_append (priv->queue, msg);
-
 	g_signal_emit (session, signals[REQUEST_QUEUED], 0, msg);
 }
 
@@ -1261,8 +1253,15 @@ static void
 cancel_message (SoupSession *session, SoupMessage *msg, guint status_code)
 {
 	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
+	SoupMessageQueueItem *item;
+
+	item = soup_message_queue_lookup (priv->queue, msg);
+	if (item) {
+		if (item->cancellable)
+			g_cancellable_cancel (item->cancellable);
+		soup_message_queue_item_unref (item);
+	}
 
-	soup_message_queue_remove_message (priv->queue, msg);
 	soup_message_io_stop (msg);
 	soup_message_set_status (msg, status_code);
 	soup_message_finished (msg);
@@ -1295,7 +1294,7 @@ gather_conns (gpointer key, gpointer host, gpointer data)
 	SoupConnection *conn = key;
 	GSList **conns = data;
 
-	*conns = g_slist_prepend (*conns, conn);
+	*conns = g_slist_prepend (*conns, g_object_ref (conn));
 }
 
 /**
@@ -1308,17 +1307,16 @@ void
 soup_session_abort (SoupSession *session)
 {
 	SoupSessionPrivate *priv;
-	SoupMessageQueueIter iter;
-	SoupMessage *msg;
+	SoupMessageQueueItem *item;
 	GSList *conns, *c;
 
 	g_return_if_fail (SOUP_IS_SESSION (session));
 	priv = SOUP_SESSION_GET_PRIVATE (session);
 
-	for (msg = soup_message_queue_first (priv->queue, &iter);
-	     msg;
-	     msg = soup_message_queue_next (priv->queue, &iter)) {
-		soup_session_cancel_message (session, msg,
+	for (item = soup_message_queue_first (priv->queue);
+	     item;
+	     item = soup_message_queue_next (priv->queue, item)) {
+		soup_session_cancel_message (session, item->msg,
 					     SOUP_STATUS_CANCELLED);
 	}
 
@@ -1327,8 +1325,6 @@ soup_session_abort (SoupSession *session)
 	conns = NULL;
 	g_hash_table_foreach (priv->conns, gather_conns, &conns);
 
-	for (c = conns; c; c = c->next)
-		g_object_ref (c->data);
 	g_mutex_unlock (priv->host_lock);
 	for (c = conns; c; c = c->next) {
 		soup_connection_disconnect (c->data);
@@ -1338,6 +1334,15 @@ soup_session_abort (SoupSession *session)
 	g_slist_free (conns);
 }
 
+/**
+ * soup_session_add_feature:
+ * @session: a #SoupSession
+ * @feature: an object that implements #SoupSessionFeature
+ *
+ * Adds @feature's functionality to @session. You can also add a
+ * feature to the session at construct time by using the
+ * %SOUP_SESSION_ADD_FEATURE property.
+ **/
 void
 soup_session_add_feature (SoupSession *session, SoupSessionFeature *feature)
 {
@@ -1349,8 +1354,22 @@ soup_session_add_feature (SoupSession *session, SoupSessionFeature *feature)
 	priv = SOUP_SESSION_GET_PRIVATE (session);
 	priv->features = g_slist_prepend (priv->features, g_object_ref (feature));
 	soup_session_feature_attach (feature, session);
+
+	if (SOUP_IS_PROXY_RESOLVER (feature))
+		priv->proxy_resolver = SOUP_PROXY_RESOLVER (feature);
 }
 
+/**
+ * soup_session_add_feature_by_type:
+ * @session: a #SoupSession
+ * @feature_type: the #GType of a class that implements #SoupSessionFeature
+ *
+ * Creates a new feature of type @feature_type and adds it to
+ * @session. You can use this instead of soup_session_add_feature() in
+ * the case wher you don't need to customize the new feature in any
+ * way. You can also add a feature to the session at construct time by
+ * using the %SOUP_SESSION_ADD_FEATURE_BY_TYPE property.
+ **/
 void
 soup_session_add_feature_by_type (SoupSession *session, GType feature_type)
 {
@@ -1364,6 +1383,13 @@ soup_session_add_feature_by_type (SoupSession *session, GType feature_type)
 	g_object_unref (feature);
 }
 
+/**
+ * soup_session_remove_feature:
+ * @session: a #SoupSession
+ * @feature: a feature that has previously been added to @session
+ *
+ * Removes @feature's functionality from @session.
+ **/
 void
 soup_session_remove_feature (SoupSession *session, SoupSessionFeature *feature)
 {
@@ -1376,9 +1402,22 @@ soup_session_remove_feature (SoupSession *session, SoupSessionFeature *feature)
 		priv->features = g_slist_remove (priv->features, feature);
 		soup_session_feature_detach (feature, session);
 		g_object_unref (feature);
+
+		if (feature == (SoupSessionFeature *)priv->proxy_resolver)
+			priv->proxy_resolver = NULL;
 	}
 }
 
+/**
+ * soup_session_remove_feature_by_type:
+ * @session: a #SoupSession
+ * @feature_type: the #GType of a class that implements #SoupSessionFeature
+ *
+ * Removes all features of type @feature_type (or any subclass of
+ * @feature_type) from @session. You can also remove standard features
+ * from the session at construct time by using the
+ * %SOUP_SESSION_REMOVE_FEATURE_BY_TYPE property.
+ **/
 void
 soup_session_remove_feature_by_type (SoupSession *session, GType feature_type)
 {
@@ -1396,3 +1435,11 @@ restart:
 		}
 	}
 }
+
+SoupProxyResolver *
+soup_session_get_proxy_resolver (SoupSession *session)
+{
+	SoupSessionPrivate *priv = SOUP_SESSION_GET_PRIVATE (session);
+
+	return priv->proxy_resolver;
+}
